package main

import "bytes"
import "compress/flate"
import "debug/elf"
import "encoding/binary"
import "io/ioutil"
import "fmt"
import "os"
import "path"
import "sort"

func must(err error) {
	if err != nil {
		panic(err)
	}
}

type ArchiveHeader struct {
	Magic, Count int32
}
type ArchiveEntry struct {
	Id, Length, Offset int32
}

var names = map[int]string{
	0: "fecs_data",
	1: "fecs_inst",
	2: "gpccs_data",
	3: "gpccs_inst",
	4: "sw_bundle_init",
	5: "sw_ctx",
	6: "sw_nonctx",
	7: "sw_method_init",
	8: "ctxreg_sys",
	9: "ctxreg_gpc",
	10: "ctxreg_tpc",
	11: "ctxreg_zcull_gpc",
	12: "ctxreg_pm_sys",
	13: "ctxreg_pm_gpc",
	14: "ctxreg_pm_tpc",
	15: "majorv",
	16: "buffer_size",
	17: "ctxsw_reg_base_index",
	18: "netlist_num",
	19: "ctxreg_ppc",
	20: "ctxreg_pmppc",
	21: "nvperf_ctxreg_sys",
	22: "nvperf_fbp_ctxregs",
	23: "nvperf_ctxreg_gpc",
	24: "nvperf_fbp_router",
	25: "nvperf_gpc_router",
	26: "ctxreg_pmltc",
	27: "ctxreg_pmfbpa",
	28: "swveidbundleinit",
	29: "nvperf_sys_router",
	30: "nvperf_pma",
	31: "ctxreg_pmrop",
	32: "ctxreg_pmucgpc",
}

func main() {
	kernel_f := os.Args[1]
	f, err := elf.Open(kernel_f)
	must(err)

	destdir := os.Args[2]

	rodata := f.Section(".rodata")
	rodataBytes, err := rodata.Data()
	must(err)

	rels := f.Section(".rela.rodata")
	relsData, err := rels.Data()
	must(err)
	if len(relsData) % 24 != 0 {
		panic(fmt.Errorf("Unexpected length for rela.rodata: %x\n", len(relsData)))
	}

	symbols, err := f.Symbols()
	must(err)

	b := bytes.NewReader(relsData)
  	var rela elf.Rela64
	var offsets []int64
	for b.Len() > 0 {
		binary.Read(b, f.ByteOrder, &rela)
		symNo := rela.Info >> 32

		sym := &symbols[symNo-1]
		if elf.SymType(sym.Info & 0xf) != elf.STT_SECTION ||
			f.Sections[sym.Section].Name != ".rodata" {
			// We're only looking for ro-data relocations
			continue
		}

		offsets = append(offsets, rela.Addend)
	}
	offsets = append(offsets, int64(len(rodataBytes)))

	sort.Slice(offsets, func (a, b int) bool {
		return offsets[a] < offsets[b]
	})

	// We assume these offsets are tightly packed in rodata. So
	// look at sequential entries.
	archiveCounter := 0
	wholeCounter := 0
	for i, off := range offsets {
		var prev int64
		if i > 0 {
			prev = offsets[i-1]
		}
		if off - prev > 32 {
			rodataReader := bytes.NewReader(rodataBytes[prev:off])
			c := flate.NewReader(rodataReader)
			data, err := ioutil.ReadAll(c)
			if err != nil {
				continue
			}

			var header ArchiveHeader
			var entries []ArchiveEntry
			dataReader := bytes.NewReader(data)
			err = binary.Read(dataReader, binary.LittleEndian, &header)
			if header.Magic != 0 {
				if len(data) < 1024 {
					continue
				}
				fname := path.Join(destdir, fmt.Sprintf("whole_%02d", wholeCounter))
				err = ioutil.WriteFile(fname, data, os.FileMode(0666))
				must(err)

				wholeCounter++
			} else {
				entries = make([]ArchiveEntry, header.Count)
				for i, _ := range entries {
					err = binary.Read(dataReader, binary.LittleEndian, &entries[i])
					if err != nil {
						break
					}
					if int(entries[i].Offset) < 8 + 12 * len(entries) {
						err = fmt.Errorf("Bad offset: %d", entries[i].Offset)
						break
					}
				}
				if err != nil || len(entries) == 0 {
					continue
				}
				archbase := path.Join(destdir, fmt.Sprintf("archive_%02d", archiveCounter))
				os.Mkdir(archbase, os.FileMode(0777))
				for _, entry := range entries {
					name := names[int(entry.Id)]
					if name == "" {
						name = fmt.Sprintf("unk%d", entry.Id)
					}
					fname := path.Join(archbase, name)
					err = ioutil.WriteFile(fname, data[entry.Offset:entry.Offset+entry.Length], os.FileMode(0666))
					must(err)
				}
				archiveCounter++
			}
		}
	}
}
